import subprocess
import sys
import time
import os
import re
import contextlib
from typing import Sequence, Set

from .version import __version__

DO_ENV_DOCS = (
    "https://docs.qhub.dev/en/latest/source/02_get_started/02_setup.html#digital-ocean"
)
AWS_ENV_DOCS = "https://docs.qhub.dev/en/latest/source/02_get_started/02_setup.html#amazon-web-services-aws"
GCP_ENV_DOCS = "https://docs.qhub.dev/en/latest/source/02_get_started/02_setup.html#google-cloud-platform"
AZURE_ENV_DOCS = "https://docs.qhub.dev/en/latest/source/02_get_started/02_setup.html#microsoft-azure"

qhub_image_tag = f"v{__version__}"
pip_install_qhub = f"pip install qhub=={__version__}"

QHUB_GH_BRANCH = os.environ.get("QHUB_GH_BRANCH", "")
if QHUB_GH_BRANCH:
    qhub_image_tag = QHUB_GH_BRANCH
    pip_install_qhub = (
        f"pip install https://github.com/Quansight/qhub/archive/{QHUB_GH_BRANCH}.zip"
    )


# Regex for suitable project names
namestr_regex = r"^[A-Za-z][A-Za-z\-_]*[A-Za-z]$"


@contextlib.contextmanager
def timer(logger, prefix):
    start_time = time.time()
    yield
    logger.info(f"{prefix} took {time.time() - start_time:.3f} [s]")


@contextlib.contextmanager
def change_directory(directory):
    current_directory = os.getcwd()
    os.chdir(directory)
    yield
    os.chdir(current_directory)


def run_subprocess_cmd(processargs, **kwargs):
    """Runs subprocess command with realtime stdout logging with optional line prefix."""
    if "prefix" in kwargs:
        line_prefix = f"[{kwargs['prefix']}]: ".encode("utf-8")
        kwargs.pop("prefix")
    else:
        line_prefix = b""

    strip_errors = kwargs.pop("strip_errors", False)

    process = subprocess.Popen(
        processargs, **kwargs, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    )
    for line in iter(lambda: process.stdout.readline(), b""):
        full_line = line_prefix + line
        if strip_errors:
            full_line = full_line.decode("utf-8")
            full_line = re.sub(
                r"\x1b\[31m", "", full_line
            )  # Remove red ANSI escape code
            full_line = full_line.encode("utf-8")

        sys.stdout.buffer.write(full_line)
        sys.stdout.flush()
    return process.wait(
        timeout=10
    )  # Should already have finished because we have drained stdout


def check_cloud_credentials(config):
    if config["provider"] == "gcp":
        for variable in {"GOOGLE_CREDENTIALS"}:
            if variable not in os.environ:
                raise ValueError(
                    f"""Missing the following required environment variable: {variable}\n
                    Please see the documentation for more information: {GCP_ENV_DOCS}"""
                )
    elif config["provider"] == "azure":
        for variable in {
            "ARM_CLIENT_ID",
            "ARM_CLIENT_SECRET",
            "ARM_SUBSCRIPTION_ID",
            "ARM_TENANT_ID",
        }:
            if variable not in os.environ:
                raise ValueError(
                    f"""Missing the following required environment variable: {variable}\n
                    Please see the documentation for more information: {AZURE_ENV_DOCS}"""
                )
    elif config["provider"] == "aws":
        for variable in {
            "AWS_ACCESS_KEY_ID",
            "AWS_SECRET_ACCESS_KEY",
        }:
            if variable not in os.environ:
                raise ValueError(
                    f"""Missing the following required environment variable: {variable}\n
                    Please see the documentation for more information: {AWS_ENV_DOCS}"""
                )
    elif config["provider"] == "do":
        for variable in {
            "AWS_ACCESS_KEY_ID",
            "AWS_SECRET_ACCESS_KEY",
            "SPACES_ACCESS_KEY_ID",
            "SPACES_SECRET_ACCESS_KEY",
            "DIGITALOCEAN_TOKEN",
        }:
            if variable not in os.environ:
                raise ValueError(
                    f"""Missing the following required environment variable: {variable}\n
                    Please see the documentation for more information: {DO_ENV_DOCS}"""
                )

        if os.environ["AWS_ACCESS_KEY_ID"] != os.environ["SPACES_ACCESS_KEY_ID"]:
            raise ValueError(
                f"""The environment variables AWS_ACCESS_KEY_ID and SPACES_ACCESS_KEY_ID must be equal\n
                See {DO_ENV_DOCS} for more information"""
            )

        if (
            os.environ["AWS_SECRET_ACCESS_KEY"]
            != os.environ["SPACES_SECRET_ACCESS_KEY"]
        ):
            raise ValueError(
                f"""The environment variables AWS_SECRET_ACCESS_KEY and SPACES_SECRET_ACCESS_KEY must be equal\n
                See {DO_ENV_DOCS} for more information"""
            )
    elif config["provider"] == "local":
        pass
    else:
        raise ValueError("Cloud Provider configuration not supported")


def check_for_duplicates(users: Sequence[dict]) -> Set:
    uids = set([])
    for user, attrs in users.items():
        if attrs["uid"] in uids:
            raise TypeError(f"Found duplicate uid ({attrs['uid']}) for {user}.")
        else:
            uids.add(attrs["uid"])
    return users



import logging

from qhub.utils import timer, check_cloud_credentials
from qhub.provider import terraform
from qhub.state import terraform_state_sync

logger = logging.getLogger(__name__)


def destroy_configuration(config, skip_remote_state_provision=False):
    logger.info(
        """Removing all infrastructure, your local files will still remain, \n
    you can use 'qhub deploy' to re - install infrastructure using same config file"""
    )

    with timer(logger, "destroying QHub"):
        # 01 Check Environment Variables
        check_cloud_credentials(config)

        # 02 Remove all infrastructure
        terraform.init(directory="infrastructure")
        terraform.refresh(directory="infrastructure")
        terraform.destroy(directory="infrastructure")

        # 03 Remove terraform backend remote state bucket
        # backwards compatible with `qhub-config.yaml` which
        # don't have `terraform_state` key
        if (
            (not skip_remote_state_provision)
            and (config.get("terraform_state", {}).get("type", "") == "remote")
            and (config.get("provider") != "local")
        ):
            terraform_state_sync(config)
            terraform.destroy(directory="terraform-state")



c.KubeController.address = ":8000"
c.KubeController.api_url = 'http://${gateway_service_name}.${gateway_service_namespace}:8000/api'
c.KubeController.gateway_instance = '${gateway_service_name}'
c.KubeController.proxy_prefix = "${gateway.prefix}"
c.KubeController.proxy_web_middlewares = [
  {
      "name": '${gateway_cluster_middleware_name}',
      "namespace": '${gateway_cluster_middleware_namespace}'
  }
]
c.KubeController.log_level = "${controller.loglevel}"
c.KubeController.completed_cluster_max_age = ${controller.completedClusterMaxAge}
c.KubeController.completed_cluster_cleanup_period = ${controller.completedClusterCleanupPeriod}
c.KubeController.backoff_base_delay = ${controller.backoffBaseDelay}
c.KubeController.backoff_max_delay = ${controller.backoffMaxDelay }
c.KubeController.k8s_api_rate_limit = ${controller.k8sApiRateLimit }
c.KubeController.k8s_api_rate_limit_burst = ${controller.k8sApiRateLimitBurst}

c.KubeController.proxy_web_entrypoint = "websecure"
c.KubeController.proxy_tcp_entrypoint = "tcp"


# override default make_ingressroute to add tls cert resolver
def make_ingressroute(self, cluster_name, namespace):
    route = f"{self.proxy_prefix}/clusters/{namespace}.{cluster_name}/"
    return {
        "apiVersion": "traefik.containo.us/v1alpha1",
        "kind": "IngressRoute",
        "metadata": {
            "labels": self.get_labels(cluster_name, "dask-scheduler"),
            "annotations": self.common_annotations,
            "name": f"dask-{cluster_name}",
        },
        "spec": {
            "entryPoints": [self.proxy_web_entrypoint],
            "routes": [
                {
                    "kind": "Rule",
                    "match": f"PathPrefix(`{route}`)",
                    "services": [
                        {
                            "name": self.make_service_name(cluster_name),
                            "namespace": namespace,
                            "port": 8787,
                        }
                    ],
                    "middlewares": self.proxy_web_middlewares,
                }
            ],
            "tls": {
                "certResolver": "default"
            }
        },
    }


c.KubeController.make_ingressroute = make_ingressroute



import pathlib

from ruamel import yaml

from qhub.render import render_template
from qhub.schema import verify


def create_render_subcommand(subparser):
    subparser = subparser.add_parser("render")
    subparser.add_argument("-o", "--output", default="./", help="output directory")
    subparser.add_argument(
        "-c", "--config", help="qhub configuration yaml file", required=True
    )
    subparser.set_defaults(func=handle_render)


def handle_render(args):
    config_filename = pathlib.Path(args.config)
    if not config_filename.is_file():
        raise ValueError(
            f"passed in configuration filename={config_filename} must exist"
        )

    with config_filename.open() as f:
        config = yaml.safe_load(f.read())

    verify(config)

    render_template(args.output, args.config, force=True)



import os
import json

c.DaskGateway.log_level = "${gateway.loglevel}"

# Configure addresses
c.DaskGateway.address = ":8000"
c.KubeBackend.api_url = 'http://{gateway_service_name}.{gateway_service_namespace}:8000/api'

c.DaskGateway.backend_class = "dask_gateway_server.backends.kubernetes.KubeBackend"
c.KubeBackend.gateway_instance = "${gateway_service_name}"

# ========= Authentication ==========
c.DaskGateway.authenticator_class = "dask_gateway_server.auth.JupyterHubAuthenticator"
c.JupyterHubAuthenticator.jupyterhub_api_url = "${jupyterhub_api_url}"

# ========= Dask Cluster Default Configuration =========
c.KubeClusterConfig.image = "${cluster-image.name}:${cluster-image.tag}"
c.KubeClusterConfig.image_pull_policy = "${cluster.image_pull_policy}"
c.KubeClusterConfig.environment = json.loads("${jsonencode(cluster.environment)}")

c.KubeClusterConfig.scheduler_cores = ${cluster.scheduler_cores}
c.KubeClusterConfig.scheduler_cores_limit = ${cluster.scheduler_cores_limit}
c.KubeClusterConfig.scheduler_memory = "${cluster.scheduler_memory}"
c.KubeClusterConfig.scheduler_memory_limit = "${cluster.scheduler_memory_limit}"
c.KubeClusterConfig.scheduler_extra_container_config = json.loads('${jsonencode(cluster.scheduler_extra_container_config)}')
c.KubeClusterConfig.scheduler_extra_pod_config = json.loads('${jsonencode(cluster.scheduler_extra_pod_config)}')

c.KubeClusterConfig.worker_cores = ${cluster.worker_cores}
c.KubeClusterConfig.worker_cores_limit = ${cluster.worker_cores_limit}
c.KubeClusterConfig.worker_memory = "${cluster.worker_memory}"
c.KubeClusterConfig.worker_memory_limit = "${cluster.worker_memory_limit}"
c.KubeClusterConfig.worker_extra_container_config = json.loads('${jsonencode(cluster.worker_extra_container_config)}')
c.KubeClusterConfig.worker_extra_pod_config = json.loads('${jsonencode(cluster.worker_extra_pod_config)}')

# ==========================================
# ========== EXTRA CONFIGURATION ===========
# ==========================================
${extra_config}



import os
import shutil

from ruamel import yaml

PROJECT_DIRECTORY = os.path.realpath(os.path.curdir)
PROVIDER = "{{ cookiecutter.provider }}"
CI_PROVIDER = "{{ cookiecutter.ci_cd.type | default('none') }}"
ENVIRONMENTS = eval("{{ cookiecutter.environments }}")
TERRAFORM_STATE = "{{ cookiecutter.terraform_state.type }}"


def remove_directory(dirpath):
    shutil.rmtree(os.path.join(PROJECT_DIRECTORY, dirpath))


def remove_file(filepath):
    os.remove(os.path.join(PROJECT_DIRECTORY, filepath))


if __name__ == "__main__":
    if ENVIRONMENTS:
        os.makedirs("environments", exist_ok=True)
        for name, spec in ENVIRONMENTS.items():
            with open(f"environments/{name}", "w") as f:
                yaml.dump(spec, f)

    # Remove any unused cloud infrastructure
    if PROVIDER != "aws":
        remove_file("infrastructure/aws.tf")
        remove_directory("infrastructure/modules/aws")

    if PROVIDER != "do":
        remove_file("infrastructure/do.tf")
        remove_directory("infrastructure/modules/digitalocean")

    if PROVIDER != "gcp":
        remove_file("infrastructure/gcp.tf")
        remove_directory("infrastructure/modules/gcp")

    if PROVIDER != "azure":
        remove_file("infrastructure/azure.tf")
        remove_directory("infrastructure/modules/azure")

    # if PROVIDER == "local" all above will have been removed

    # Remove any unused state

    if TERRAFORM_STATE == "local":
        remove_file("infrastructure/state.tf")

    if TERRAFORM_STATE != "remote" or PROVIDER == "local":
        remove_directory("terraform-state")
    else:
        # Remove the clouds we don't need:
        if PROVIDER != "aws":
            remove_directory("terraform-state/modules/aws")

        if PROVIDER != "do":
            remove_directory("terraform-state/modules/digitalocean")

        if PROVIDER != "gcp":
            remove_directory("terraform-state/modules/gcp")

        if PROVIDER != "azure":
            remove_directory("terraform-state/modules/azure")

    # Remove any unused CI

    if CI_PROVIDER != "github-actions":
        remove_directory(".github")

    if CI_PROVIDER != "gitlab-ci":
        remove_file(".gitlab-ci.yml")

    # if CI_PROVIDER == "none" all above will have been removed

    # templates directory is only used by includes
    remove_directory("templates")



import enum
import typing
from abc import ABC

import pydantic
from pydantic import validator, root_validator
from qhub.utils import namestr_regex, check_for_duplicates


class CertificateEnum(str, enum.Enum):
    letsencrypt = "lets-encrypt"
    selfsigned = "self-signed"
    existing = "existing"


class TerraformStateEnum(str, enum.Enum):
    remote = "remote"
    local = "local"
    existing = "existing"


class ProviderEnum(str, enum.Enum):
    local = "local"
    do = "do"
    aws = "aws"
    gcp = "gcp"
    azure = "azure"


class CiEnum(str, enum.Enum):
    github_actions = "github-actions"
    gitlab_ci = "gitlab-ci"
    none = "none"


class AuthenticationEnum(str, enum.Enum):
    password = "password"
    github = "GitHub"
    auth0 = "Auth0"
    custom = "custom"


class Base(pydantic.BaseModel):
    ...

    class Config:
        extra = "forbid"


# ============== CI/CD =============


class CICD(Base):
    type: CiEnum
    branch: str
    before_script: typing.Optional[typing.List[str]]
    after_script: typing.Optional[typing.List[str]]


# ============== Monitoring =============


class Monitoring(Base):
    enabled: bool


# ============== ClearML =============


class ClearML(Base):
    enabled: bool
    enable_forward_auth: typing.Optional[bool]


# ============== Prefect =============


class Prefect(Base):
    enabled: bool
    image: typing.Optional[str]
    overrides: typing.Optional[typing.Dict]


# ============= Terraform ===============


class TerraformState(Base):
    type: TerraformStateEnum
    backend: typing.Optional[str]
    config: typing.Optional[typing.Dict[str, str]]


class TerraformModules(Base):
    # No longer used, so ignored, but could still be in qhub-config.yaml
    repository: str
    rev: str


# ============ Certificate =============


class Certificate(Base):
    type: CertificateEnum
    # existing
    secret_name: typing.Optional[str]
    # lets-encrypt
    acme_email: typing.Optional[str]
    acme_server: typing.Optional[str]


# ========== Default Images ==============


class DefaultImages(Base):
    jupyterhub: str
    jupyterlab: str
    dask_worker: str
    dask_gateway: str
    conda_store: str


# =========== Authentication ==============


class GitHubConfig(Base):
    oauth_callback_url: str
    client_id: str
    client_secret: str


class Auth0Config(Base):
    client_id: str
    client_secret: str
    oauth_callback_url: str
    scope: typing.List[str]
    auth0_subdomain: str


class Authentication(Base, ABC):
    _types: typing.Dict[str, type] = {}

    type: AuthenticationEnum

    # Based on https://github.com/samuelcolvin/pydantic/issues/2177#issuecomment-739578307

    # This allows type field to determine which subclass of Authentication should be used for validation.

    # Used to register automatically all the submodels in `_types`.
    def __init_subclass__(cls):
        cls._types[cls._typ.value] = cls

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value: typing.Dict[str, typing.Any]) -> "Authentication":
        if "type" not in value:
            raise ValueError("type field is missing from security.authentication")

        specified_type = value.get("type")
        sub_class = cls._types.get(specified_type, None)

        if not sub_class:
            raise ValueError(
                f"No registered Authentication type called {specified_type}"
            )

        # init with right submodel
        return sub_class(**value)


class PasswordAuthentication(Authentication):
    _typ = AuthenticationEnum.password


class Auth0Authentication(Authentication):
    _typ = AuthenticationEnum.auth0
    config: Auth0Config


class GitHubAuthentication(Authentication):
    _typ = AuthenticationEnum.github
    config: GitHubConfig


class CustomAuthentication(Authentication):
    _typ = AuthenticationEnum.custom
    authentication_class: str
    config: typing.Dict[str, typing.Any]


# =========== Users and Groups =============


class User(Base):
    uid: str
    password: typing.Optional[str]
    primary_group: str
    secondary_groups: typing.Optional[typing.List[str]]


class Group(Base):
    gid: int


# ============== Security ================


class Security(Base):
    authentication: Authentication
    users: typing.Dict[str, User]
    groups: typing.Dict[str, Group]

    @validator("users", pre=True)
    def validate_uderids(cls, v):
        # raise TypeError if duplicated
        return check_for_duplicates(v)


# ================ Providers ===============


class KeyValueDict(Base):
    key: str
    value: str


class NodeSelector(Base):
    general: KeyValueDict
    user: KeyValueDict
    worker: KeyValueDict


class NodeGroup(Base):
    instance: str
    min_nodes: int
    max_nodes: int
    gpu: typing.Optional[bool] = False

    class Config:
        extra = "allow"


class DigitalOceanProvider(Base):
    region: str
    kubernetes_version: str
    node_groups: typing.Dict[str, NodeGroup]


class GoogleCloudPlatformProvider(Base):
    project: str
    region: str
    zone: typing.Optional[str]  # No longer used
    availability_zones: typing.Optional[typing.List[str]]  # Genuinely optional
    kubernetes_version: str
    node_groups: typing.Dict[str, NodeGroup]


class AzureProvider(Base):
    project: str
    region: str
    kubernetes_version: str
    node_groups: typing.Dict[str, NodeGroup]
    storage_account_postfix: str


class AmazonWebServicesProvider(Base):
    region: str
    availability_zones: typing.Optional[typing.List[str]]
    kubernetes_version: str
    node_groups: typing.Dict[str, NodeGroup]


class LocalProvider(Base):
    kube_context: typing.Optional[str]
    node_selectors: typing.Dict[str, KeyValueDict]


# ================= Theme ==================


class Theme(Base):
    jupyterhub: typing.Dict[str, str]


# ================== Profiles ==================


class KubeSpawner(Base):
    cpu_limit: int
    cpu_guarantee: int
    mem_limit: str
    mem_guarantee: str
    image: str

    class Config:
        extra = "allow"


class JupyterLabProfile(Base):
    display_name: str
    description: str
    default: typing.Optional[bool]
    users: typing.Optional[typing.List[str]]
    groups: typing.Optional[typing.List[str]]
    kubespawner_override: typing.Optional[KubeSpawner]


class DaskWorkerProfile(Base):
    worker_cores_limit: int
    worker_cores: int
    worker_memory_limit: str
    worker_memory: str
    image: str

    class Config:
        extra = "allow"


class Profiles(Base):
    jupyterlab: typing.List[JupyterLabProfile]
    dask_worker: typing.Dict[str, DaskWorkerProfile]

    @validator("jupyterlab", pre=True)
    def check_default(cls, v, values):
        """Check if only one default value is present"""
        default = [attrs["default"] for attrs in v if "default" in attrs]
        if default.count(True) > 1:
            raise TypeError(
                "Multiple default Jupyterlab profiles may cause unexpected problems."
            )
        return v


# ================ Environment ================


class CondaEnvironment(Base):
    name: str
    channels: typing.Optional[typing.List[str]]
    dependencies: typing.List[typing.Union[str, typing.Dict[str, typing.List[str]]]]


# =============== CDSDashboards ==============


class CDSDashboards(Base):
    enabled: bool
    cds_hide_user_named_servers: typing.Optional[bool]
    cds_hide_user_dashboard_servers: typing.Optional[bool]


# ======== External Container Registry ========

# This allows the user to set a private AWS ECR as a replacement for
# Docker Hub for some images - those where you provide the full path
# to the image on the ECR.
# extcr_account and extcr_region are the AWS account number and region
# of the ECR respectively. access_key_id and secret_access_key are
# AWS access keys that should have read access to the ECR.


class ExtContainerReg(Base):
    enabled: bool
    access_key_id: typing.Optional[str]
    secret_access_key: typing.Optional[str]
    extcr_account: typing.Optional[str]
    extcr_region: typing.Optional[str]

    @root_validator
    def enabled_must_have_fields(cls, values):
        if values["enabled"]:
            for fldname in (
                "access_key_id",
                "secret_access_key",
                "extcr_account",
                "extcr_region",
            ):
                if (
                    fldname not in values
                    or values[fldname] is None
                    or values[fldname].strip() == ""
                ):
                    raise ValueError(
                        f"external_container_reg must contain a non-blank {fldname} when enabled is true"
                    )
        return values


# ==================== Main ===================

letter_dash_underscore_pydantic = pydantic.constr(regex=namestr_regex)


class Main(Base):
    project_name: letter_dash_underscore_pydantic
    namespace: typing.Optional[letter_dash_underscore_pydantic]
    provider: ProviderEnum
    ci_cd: typing.Optional[CICD]
    domain: str
    terraform_state: typing.Optional[TerraformState]
    terraform_modules: typing.Optional[
        TerraformModules
    ]  # No longer used, so ignored, but could still be in qhub-config.yaml
    certificate: Certificate
    prefect: typing.Optional[Prefect]
    cdsdashboards: CDSDashboards
    security: Security
    external_container_reg: typing.Optional[ExtContainerReg]
    default_images: DefaultImages
    storage: typing.Dict[str, str]
    local: typing.Optional[LocalProvider]
    google_cloud_platform: typing.Optional[GoogleCloudPlatformProvider]
    amazon_web_services: typing.Optional[AmazonWebServicesProvider]
    azure: typing.Optional[AzureProvider]
    digital_ocean: typing.Optional[DigitalOceanProvider]
    theme: Theme
    profiles: Profiles
    environments: typing.Dict[str, CondaEnvironment]
    monitoring: typing.Optional[Monitoring]
    clearml: typing.Optional[ClearML]


def verify(config):
    Main(**config)



import pytest

from ruamel.yaml import YAML

from qhub.render import render_template, set_env_vars_in_config, remove_existing_renders
from qhub.initialize import render_config

INIT_INPUTS = [
    # project, namespace, domain, cloud_provider, ci_provider, auth_provider
    ("do-pytest", "dev", "do.qhub.dev", "do", "github-actions", "github"),
    ("aws-pytest", "dev", "aws.qhub.dev", "aws", "github-actions", "github"),
    ("gcp-pytest", "dev", "gcp.qhub.dev", "gcp", "github-actions", "github"),
    ("azure-pytest", "dev", "azure.qhub.dev", "azure", "github-actions", "github"),
]
QHUB_CONFIG_FN = "qhub-config.yaml"
PRESERVED_DIR = "preserved_dir"


@pytest.fixture
def init_render(request, tmp_path):
    (
        project,
        namespace,
        domain,
        cloud_provider,
        ci_provider,
        auth_provider,
    ) = request.param

    output_directory = tmp_path / f"{cloud_provider}_output_dir"
    output_directory.mkdir()
    qhub_config = output_directory / QHUB_CONFIG_FN

    # data that should NOT be deleted when `qhub render` is called
    preserved_directory = output_directory / PRESERVED_DIR
    preserved_directory.mkdir()
    preserved_filename = preserved_directory / "file.txt"
    preserved_filename.write_text("This is a test...")

    config = render_config(
        project_name=project,
        namespace=namespace,
        qhub_domain=domain,
        cloud_provider=cloud_provider,
        ci_provider=ci_provider,
        repository="github.com/test/test",
        auth_provider=auth_provider,
        repository_auto_provision=False,
        auth_auto_provision=False,
        terraform_state="remote",
        kubernetes_version="1.18.0",
        disable_prompt=True,
    )
    yaml = YAML(typ="unsafe", pure=True)
    yaml.dump(config, qhub_config)

    render_template(str(output_directory), qhub_config, force=True)

    yield (output_directory, request)


def test_get_secret_config_entries(monkeypatch):
    sec1 = "secret1"
    sec2 = "nestedsecret1"
    config_orig = {
        "key1": "value1",
        "key2": "QHUB_SECRET_secret_val",
        "key3": {
            "nested_key1": "nested_value1",
            "nested_key2": "QHUB_SECRET_nested_secret_val",
        },
    }
    expected = {
        "key1": "value1",
        "key2": sec1,
        "key3": {
            "nested_key1": "nested_value1",
            "nested_key2": sec2,
        },
    }

    # should raise error if implied env var is not set
    with pytest.raises(EnvironmentError):
        config = config_orig.copy()
        set_env_vars_in_config(config)

    monkeypatch.setenv("secret_val", sec1, prepend=False)
    monkeypatch.setenv("nested_secret_val", sec2, prepend=False)
    config = config_orig.copy()
    set_env_vars_in_config(config)
    assert config == expected


@pytest.mark.parametrize(
    "init_render",
    INIT_INPUTS,
    indirect=True,
)
def test_render_template(init_render):
    output_directory, request = init_render
    (
        project,
        namespace,
        domain,
        cloud_provider,
        ci_provider,
        auth_provider,
    ) = request.param
    qhub_config = output_directory / QHUB_CONFIG_FN

    yaml = YAML()
    qhub_config_json = yaml.load(qhub_config.read_text())

    assert qhub_config_json["project_name"] == project
    assert qhub_config_json["namespace"] == namespace
    assert qhub_config_json["domain"] == domain
    assert qhub_config_json["provider"] == cloud_provider


@pytest.mark.parametrize(
    "init_render",
    INIT_INPUTS,
    indirect=True,
)
def test_remove_existing_renders(init_render):
    output_directory, request = init_render
    dirs = [_.name for _ in output_directory.iterdir()]
    preserved_files = [_ for _ in (output_directory / PRESERVED_DIR).iterdir()]

    # test `remove_existing_renders` implicitly
    assert PRESERVED_DIR in dirs
    assert len(preserved_files[0].read_text()) > 0

    # test `remove_existing_renders` explicitly
    remove_existing_renders(output_directory)

    assert PRESERVED_DIR in dirs
    assert len(preserved_files[0].read_text()) > 0



import pathlib
import collections
import functools
import json
import os
from shutil import rmtree

from ruamel import yaml
from cookiecutter.generate import generate_files
from ..version import __version__
from ..constants import TERRAFORM_VERSION
from ..utils import pip_install_qhub, QHUB_GH_BRANCH


def patch_dask_gateway_extra_config(config):
    """Basically the current dask_gateway helm chart only allows one
    update to extraContainerConfig and extraPodConfig for the workers
    and scheduler. Thus we need to copy the configuration done in
    these settings. The only critical one is mounting the conda store
    directory.

    """
    namespace = config["namespace"]
    conda_store_volume = {
        "name": "conda-store",
        "persistentVolumeClaim": {"claimName": f"conda-store-{namespace}-share"},
    }
    extra_pod_config = {"volumes": [conda_store_volume]}

    merge_config_for = ["worker_extra_pod_config", "scheduler_extra_pod_config"]

    if "profiles" in config and "dask_worker" in config["profiles"]:
        for worker_name, worker_config in config["profiles"]["dask_worker"].items():
            for config_name in merge_config_for:
                if config_name in worker_config:
                    worker_config[config_name] = deep_merge(
                        worker_config[config_name], extra_pod_config
                    )


def patch_versioning_extra_config(config):
    """
    Set defaults for qhub_version and pip install command
    because they depend on __version__ so cannot be static in cookiecutter.json
    """
    if "qhub_version" not in config:
        config["qhub_version"] = __version__

    config["pip_install_qhub"] = pip_install_qhub

    config["QHUB_GH_BRANCH"] = QHUB_GH_BRANCH

    if "terraform_version" not in config:
        config["terraform_version"] = TERRAFORM_VERSION


def deep_merge(d1, d2):
    """Deep merge two dictionaries.
    >>> value_1 = {
    'a': [1, 2],
    'b': {'c': 1, 'z': [5, 6]},
    'e': {'f': {'g': {}}},
    'm': 1,
    }

    >>> value_2 = {
        'a': [3, 4],
        'b': {'d': 2, 'z': [7]},
        'e': {'f': {'h': 1}},
        'm': [1],
    }

    >>> print(deep_merge(value_1, value_2))
    {'m': 1, 'e': {'f': {'g': {}, 'h': 1}}, 'b': {'d': 2, 'c': 1, 'z': [5, 6, 7]}, 'a': [1, 2, 3,  4]}
    """
    if isinstance(d1, dict) and isinstance(d2, dict):
        d3 = {}
        for key in d1.keys() | d2.keys():
            if key in d1 and key in d2:
                d3[key] = deep_merge(d1[key], d2[key])
            elif key in d1:
                d3[key] = d1[key]
            elif key in d2:
                d3[key] = d2[key]
        return d3
    elif isinstance(d1, list) and isinstance(d2, list):
        return [*d1, *d2]
    else:  # if they don't match use left one
        return d1


def render_template(output_directory, config_filename, force=False):
    import qhub

    input_directory = pathlib.Path(qhub.__file__).parent / "template"

    # would be nice to remove assumption that input directory
    # is in local filesystem
    input_directory = pathlib.Path(input_directory)
    if not input_directory.is_dir():
        raise ValueError(f"input directory={input_directory} is not a directory")

    output_directory = pathlib.Path(output_directory).resolve()
    # due to cookiecutter requiring a template directory folder
    # we take the output directory and split into two components
    repo_directory = output_directory.name
    output_directory = output_directory.parent

    # mkdir all the way down to repo dir so we can copy .gitignore into it in remove_existing_renders
    (output_directory / repo_directory).mkdir(exist_ok=True, parents=True)

    filename = pathlib.Path(config_filename)

    if not filename.is_file():
        raise ValueError(f"cookiecutter configuration={filename} is not filename")

    with filename.open() as f:
        config = yaml.safe_load(f)

        # For any config values that start with
        # QHUB_SECRET_, set the values using the
        # corresponding env var.
        set_env_vars_in_config(config)

        config["repo_directory"] = repo_directory
        patch_dask_gateway_extra_config(config)

    with (input_directory / "cookiecutter.json").open() as f:
        config = collections.ChainMap(config, json.load(f))

    patch_versioning_extra_config(config)

    remove_existing_renders(
        dest_repo_dir=output_directory / repo_directory,
        verbosity=2,
    )

    generate_files(
        repo_dir=str(input_directory),
        context={"cookiecutter": config},
        output_dir=str(output_directory),
        overwrite_if_exists=force,
    )


def remove_existing_renders(dest_repo_dir, verbosity=0):
    """
    Remove all files and directories beneath each directory in `deletable_dirs`. These files and directories will be regenerated in the next step (`generate_files`) based on the configurations set in `qhub-config.yml`.

    Inputs must be pathlib.Path
    """
    home_dir = pathlib.Path.home()
    if pathlib.Path.cwd() == home_dir:
        raise ValueError(
            f"Deploying QHub from the home directory, {home_dir}, is not permitted."
        )

    deletable_dirs = [
        "terraform-state",
        ".github",
        "infrastructure",
        "image",
        ".gitlab-ci.yml",
    ]

    for deletable_dir in deletable_dirs:
        deletable_dir = dest_repo_dir / deletable_dir
        if deletable_dir.exists():
            if verbosity > 0:
                print(f"Deleting all files and directories beneath {deletable_dir} ...")
            rmtree(deletable_dir)


def set_env_vars_in_config(config):
    """

    For values in the config starting with 'QHUB_SECRET_XXX' the environment
    variables are searched for the pattern XXX and the config value is
    modified. This enables setting secret values that should not be directly
    stored in the config file.

    NOTE: variables are most likely written to a file somewhere upon render. In
    order to further reduce risk of exposure of any of these variables you might
    consider preventing storage of the terraform render output.
    """
    private_entries = get_secret_config_entries(config)
    for idx in private_entries:
        set_qhub_secret(config, idx)


def get_secret_config_entries(config, config_idx=None, private_entries=None):
    output = private_entries or []
    if config_idx is None:
        sub_dict = config
        config_idx = []
    else:
        sub_dict = get_sub_config(config, config_idx)

    for key, value in sub_dict.items():
        if type(value) is dict:
            sub_dict_outputs = get_secret_config_entries(
                config, [*config_idx, key], private_entries
            )
            output = [*output, *sub_dict_outputs]
        else:
            if "QHUB_SECRET_" in str(value):
                output = [*output, [*config_idx, key]]
    return output


def get_sub_config(conf, conf_idx):
    sub_config = functools.reduce(dict.__getitem__, conf_idx, conf)
    return sub_config


def set_sub_config(conf, conf_idx, value):

    get_sub_config(conf, conf_idx[:-1])[conf_idx[-1]] = value


def set_qhub_secret(config, idx):
    placeholder = get_sub_config(config, idx)
    secret_var = get_qhub_secret(placeholder)
    set_sub_config(config, idx, secret_var)


def get_qhub_secret(secret_var):
    env_var = secret_var.lstrip("QHUB_SECRET_")
    val = os.environ.get(env_var)
    if not val:
        raise EnvironmentError(
            f"Since '{secret_var}' was found in the"
            " QHub config, the environment variable"
            f" '{env_var}' must be set."
        )
    return val



import logging
import os
import re
from subprocess import CalledProcessError

from qhub.provider import terraform
from qhub.utils import timer, check_cloud_credentials
from qhub.provider.dns.cloudflare import update_record
from qhub.state import terraform_state_sync

logger = logging.getLogger(__name__)


def deploy_configuration(
    config,
    dns_provider,
    dns_auto_provision,
    disable_prompt,
    skip_remote_state_provision,
):
    logger.info(f'All qhub endpoints will be under https://{config["domain"]}')

    with timer(logger, "deploying QHub"):
        try:
            guided_install(
                config,
                dns_provider,
                dns_auto_provision,
                disable_prompt,
                skip_remote_state_provision,
            )
        except CalledProcessError as e:
            logger.error(e.output)
            raise e


def guided_install(
    config,
    dns_provider,
    dns_auto_provision,
    disable_prompt=False,
    skip_remote_state_provision=False,
):
    # 01 Check Environment Variables
    check_cloud_credentials(config)
    # Check that secrets required for terraform
    # variables are set as required
    check_secrets(config)

    # 02 Create terraform backend remote state bucket
    # backwards compatible with `qhub-config.yaml` which
    # don't have `terraform_state` key
    if (
        (not skip_remote_state_provision)
        and (config.get("terraform_state", {}).get("type", "") == "remote")
        and (config.get("provider") != "local")
    ):
        terraform_state_sync(config)

    # 3 kuberentes-alpha provider requires that kubernetes be
    # provisionioned before any "kubernetes_manifests" resources
    terraform.init(directory="infrastructure")
    terraform.apply(
        directory="infrastructure",
        targets=[
            "module.kubernetes",
            "module.kubernetes-initialization",
        ],
    )

    # 04 Create qhub initial state (up to nginx-ingress)
    terraform.init(directory="infrastructure")
    terraform.apply(
        directory="infrastructure",
        targets=[
            "module.kubernetes",
            "module.kubernetes-initialization",
            "module.kubernetes-ingress",
        ],
    )

    cmd_output = terraform.output(directory="infrastructure")
    # This is a bit ugly, but the issue we have at the moment is being unable
    # to parse cmd_output as json on Github Actions.
    ip_matches = re.findall(r'"ip": "(?!string)(.+)"', cmd_output)
    hostname_matches = re.findall(r'"hostname": "(?!string)(.+)"', cmd_output)
    if ip_matches:
        ip_or_hostname = ip_matches[0]
    elif hostname_matches:
        ip_or_hostname = hostname_matches[0]
    else:
        raise ValueError(f"IP Address not found in: {cmd_output}")

    # 05 Update DNS to point to qhub deployment
    if dns_auto_provision and dns_provider == "cloudflare":
        record_name, zone_name = (
            config["domain"].split(".")[:-2],
            config["domain"].split(".")[-2:],
        )
        record_name = ".".join(record_name)
        zone_name = ".".join(zone_name)
        if config["provider"] in {"do", "gcp", "azure"}:
            update_record(zone_name, record_name, "A", ip_or_hostname)
            if config.get("clearml", {}).get("enabled"):
                add_clearml_dns(zone_name, record_name, "A", ip_or_hostname)
        elif config["provider"] == "aws":
            update_record(zone_name, record_name, "CNAME", ip_or_hostname)
            if config.get("clearml", {}).get("enabled"):
                add_clearml_dns(zone_name, record_name, "CNAME", ip_or_hostname)
        else:
            logger.info(
                f"Couldn't update the DNS record for cloud provider: {config['provider']}"
            )
    elif not disable_prompt:
        input(
            f"Take IP Address {ip_or_hostname} and update DNS to point to "
            f'"{config["domain"]}" [Press Enter when Complete]'
        )

    # 06 Full deploy QHub
    terraform.apply(directory="infrastructure")


def add_clearml_dns(zone_name, record_name, record_type, ip_or_hostname):
    logger.info(f"Setting DNS record for ClearML for record: {record_name}")
    dns_records = [
        f"app.clearml.{record_name}",
        f"api.clearml.{record_name}",
        f"files.clearml.{record_name}",
    ]

    for dns_record in dns_records:
        update_record(zone_name, dns_record, record_type, ip_or_hostname)


def check_secrets(config):
    """
    Checks that the appropriate variables are set based on the current config.
    These variables are prefixed with TF_VAR_ and are used to populate the
    corresponding variables in the terraform deployment. e.g.
    TF_VAR_prefect_token sets the prefect_token variable in Terraform. These
    values are set in the terraform state but are not leaked when the
    terraform render occurs.
    """

    missing_env_vars = []

    # Check prefect integration set up.
    if "prefect" in config and config["prefect"]["enabled"]:
        var = "TF_VAR_prefect_token"
        if var not in os.environ:
            missing_env_vars.append(var)

    if missing_env_vars:
        raise EnvironmentError(
            "Some environment variables used to propagate secrets to the "
            "terraform deployment were not set. Please set these before "
            f"continuing: {', '.join(missing_env_vars)}"
        )



import os

from qhub.provider import terraform


def terraform_state_sync(config, logger=None):
    """
    Clear terraform-state terraform.tfstate and re-import from cloud (if exists)
    then terraform apply to ensure state is correct in cloud too
    """
    terraform.init(directory="terraform-state")

    provider = config.get("provider")
    project_name = config.get("project_name")
    namespace = config.get("namespace")

    terraform.rm_local_state(directory="terraform-state")

    # Example imports:

    # AWS
    # terraform import module.terraform-state.aws_s3_bucket.terraform-state qhubintgrtnaws-dev-terraform-state
    # terraform import module.terraform-state.aws_dynamodb_table.terraform-state-lock qhubintgrtnaws-dev-terraform-state-lock

    # GCP
    # terraform import module.terraform-state.module.gcs.google_storage_bucket.static-site qhubintgrtngcp-dev-terraform-state
    # But needs terraform apply to import force_destroy: https://github.com/hashicorp/terraform-provider-google/issues/1509

    try:
        if provider == "aws":
            terraform.tfimport(
                "module.terraform-state.aws_s3_bucket.terraform-state",
                f"{project_name}-{namespace}-terraform-state",
                directory="terraform-state",
            )
            terraform.tfimport(
                "module.terraform-state.aws_dynamodb_table.terraform-state-lock",
                f"{project_name}-{namespace}-terraform-state-lock",
                directory="terraform-state",
            )

        elif provider == "gcp":
            terraform.tfimport(
                "module.terraform-state.module.gcs.google_storage_bucket.static-site",
                f"{project_name}-{namespace}-terraform-state",
                directory="terraform-state",
            )

        elif provider == "do":
            do_region = config.get("digital_ocean", {}).get("region", "nyc3")

            terraform.tfimport(
                "module.terraform-state.module.spaces.digitalocean_spaces_bucket.main",
                f"{do_region},{project_name}-{namespace}-terraform-state",
                directory="terraform-state",
            )

        elif provider == "azure":
            storage_account_postfix = config.get("azure", {}).get(
                "storage_account_postfix", ""
            )

            if storage_account_postfix == "":
                raise ValueError(
                    "azure: storage_account_postfix not present in config file"
                )

            subscription_id = os.environ.get("ARM_SUBSCRIPTION_ID", "")

            if subscription_id == "":
                raise ValueError("ARM_SUBSCRIPTION_ID environment variable is not set")

            resource_group_name = f"{project_name}-{namespace}"
            resource_group_name_safe = resource_group_name.replace("-", "")

            resource_group_url = f"/subscriptions/{subscription_id}/resourceGroups/{project_name}-{namespace}"

            terraform.tfimport(
                "module.terraform-state.azurerm_resource_group.terraform-resource-group",
                resource_group_url,
                directory="terraform-state",
            )
            terraform.tfimport(
                "module.terraform-state.azurerm_storage_account.terraform-storage-account",
                f"{resource_group_url}/providers/Microsoft.Storage/storageAccounts/{resource_group_name_safe}{storage_account_postfix}",
                directory="terraform-state",
            )
            terraform.tfimport(
                "module.terraform-state.azurerm_storage_container.storage_container",
                f"https://{resource_group_name_safe}{storage_account_postfix}.blob.core.windows.net/{resource_group_name}state",
                directory="terraform-state",
            )

    except terraform.TerraformException:
        if logger:
            logger.info(
                "Terraform error presumed to be due to importing a resource which does not exist yet"
            )

    terraform.apply(
        directory="terraform-state"
    )  # If before destroy, this is mainly to sync force_destroy attribute to buckets
    # If before deploy, this may be to create the resources for the first time



import pytest
from qhub.deploy import check_secrets


def test_check_secrets(monkeypatch):

    # Should do nothing without prefect key or not enabled
    check_secrets({"key": "value"})
    check_secrets({"prefect": {"enabled": False}})

    # Should do nothing if appropriate var is set
    monkeypatch.setenv("TF_VAR_prefect_token", "secret_token")
    check_secrets({"prefect": {"enabled": True}})

    # Should raise error if var is not set
    monkeypatch.delenv("TF_VAR_prefect_token")
    with pytest.raises(EnvironmentError):
        check_secrets({"prefect": {"enabled": True}})



import os
import functools
from azure.identity import DefaultAzureCredential
from azure.mgmt.containerservice import ContainerServiceClient


@functools.lru_cache()
def initiate_container_service_client():
    subscription_id = os.environ.get("ARM_SUBSCRIPTION_ID", None)
    return ContainerServiceClient(
        credential=DefaultAzureCredential(), subscription_id=subscription_id
    )


@functools.lru_cache()
def kubernetes_versions(azure_location="Central US"):

    client = initiate_container_service_client()
    azure_location = azure_location.replace(" ", "").lower()

    k8s_versions_list = client.container_services.list_orchestrators(
        azure_location, resource_type="managedClusters"
    ).as_dict()
    supported_kubernetes_versions = []

    for key in k8s_versions_list["orchestrators"]:
        if key["orchestrator_type"] == "Kubernetes":
            supported_kubernetes_versions.append(key["orchestrator_version"])

    return supported_kubernetes_versions
